package ch.hepia.numeric;

import org.junit.jupiter.api.Test;

import java.util.ArrayList;

//import sun.jvmstat.monitor.VmIdentifier;

import java.util.List;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.assertFalse;


class VectorTest {

    @Test
    void vectorMainOperations() {
        Vector v1 = Vector.of(1.0, 2.0, 3.0);
        Vector v2 = Vector.of(-1.0, -2.0, -3.0);
        assertEquals(v1.add(v2), Vector.of(0.0, 0.0, 0.0));
        assertEquals(v1.sub(v2), Vector.of(2.0, 4.0, 6.0));
        assertEquals(v1.mul(3.0), Vector.of(3.0, 6.0, 9.0));
        assertEquals(Vector.empty().len(), 0);
        assertEquals(Vector.of(1.0, 2.0, 1.0).len(), 3);
    }
    
    @Test
    void vectorCreation() {
        assertEquals(Vector.zeros(3), Vector.of(0.0, 0.0, 0.0));
        assertEquals(Vector.ones(3), Vector.of(1.0, 1.0, 1.0));
        assertEquals(Vector.of(1.0, 2.0).map( d -> d * 2.0), Vector.of(2.0, 4.0));
        assertEquals(Vector.fill(3, 1.7), Vector.of(1.7, 1.7, 1.7));
        assertEquals(Vector.tabulate(4, i -> i*5.0), Vector.of(0.0, 5.0, 10.0, 15.0));
    }
    
    @Test
    void vectorLinespace() {
        assertEquals(Vector.linespace(0.0, 1.0, 3), Vector.of(0.0, 0.5, 1.0));
        assertEquals(Vector.linespace(0.0, 1.0, 5), Vector.of(0.0, 0.25, 0.5, 0.75, 1.0));
    }

    @Test
    void vectorNormAndSum() {
        assertEquals(Vector.of(1.0, 2.0, 2.0).norm(), 3.0);
        assertEquals(Vector.sum(List.of(Vector.of(1.0, 2.0, 1.0), Vector.of(0.0, 0.0, 1.0))).norm(), 3.0);
        assertEquals(
                Vector.sum(List.of(Vector.of(1.0, 2.0, 1.0), Vector.of(0.0, 0.0, 1.0))).norm(),
                Vector.norms(List.of(Vector.of(1.0, 2.0, 1.0), Vector.of(0.0, 0.0, 1.0)))
        );
        assertEquals(Vector.sum(List.of(Vector.of(1.0, 2.0, 1.0), Vector.of(0.0, 0.0, 1.0))), Vector.of(1.0, 2.0, 2.0));
    }
    
    @Test
    void vectorTranspose() {
        Vector v1 = Vector.of(1.0, 2.0, 3.0);
        Vector v2 = Vector.of(1.0, 0.0, -1.0);
        assertEquals(v1.t().dot(v2), -2.0);
        /*
         * v1.dot(v2) should not compile !!!
         * only a transposed vector with a vector
         */
        Vector v3 = Vector.of(1.0, 2.0, 3.0);
        Vector v4 = Vector.of(1.0, 2.0, 3.0);
        assertEquals( v3.t().dot(v4), 14.0 );
        assertEquals( v3.t().sub(v4.t()), Vector.of(0.0, 0.0, 0.0).t() );
    }

    @Test
    void vectorToString() {
        Vector v = Vector.of(1.0, 2.0, 3.0);
        assertEquals(v.toString(), "Vector[1.0, 2.0, 3.0]");
        assertEquals(v.t().toString(), "Transposed[1.0, 2.0, 3.0]");
    }

    @Test
    void vectorSlice() {
        Vector v = Vector.of(1.0, 2.0, 3.0, 4.0, 5.0);
        assertEquals(v.sliceFrom(2), Vector.of(3.0, 4.0, 5.0));
        assertEquals(v.sliceTo(3), Vector.of(1.0, 2.0, 3.0));
        assertEquals(v.slice(2,4), Vector.of(3.0, 4.0));
        assertEquals(v.slice(4,4), Vector.empty());
        assertEquals(v.slice(4,2), Vector.empty());
        assertEquals(v.removed(2), Vector.of(1.0, 2.0, 4.0, 5.0));
    }
    @Test
    void vectorConcat() {
        Vector v1 = Vector.of(1.0, 2.0, 3.0);
        Vector v2 = Vector.of(-1.0, -2.0, -3.0);
        assertEquals(v1.concat(v2), Vector.of(1.0, 2.0, 3.0, -1.0, -2.0, -3.0));
    }
// Test Transposed :
    @Test
    void transposedMainOp() {
        Transposed v1 = Transposed.of(1.0, 2.0, 3.0);
        Transposed v2 = Transposed.of(-1.0, -2.0, -3.0);
        assertEquals(v1.add(v2), Transposed.of(0.0, 0.0, 0.0));
        assertEquals(v1.sub(v2), Transposed.of(2.0, 4.0, 6.0));
        assertEquals(v1.mul(3.0), Transposed.of(3.0, 6.0, 9.0));
        assertEquals(Transposed.empty().len(), 0);
        assertEquals(Transposed.of(1.0, 2.0, 1.0).len(), 3);
    }

    @Test
    void transposedNormAndSum() {
        assertEquals(Transposed.of(1.0, 2.0, 2.0).norm(), 3.0);
        assertEquals(Transposed.sum(List.of(Transposed.of(1.0, 2.0, 1.0), Transposed.of(0.0, 0.0, 1.0))).norm(), 3.0);
        assertEquals(
                Transposed.sum(List.of(Transposed.of(1.0, 2.0, 1.0), Transposed.of(0.0, 0.0, 1.0))).norm(),
                Transposed.norms(List.of(Transposed.of(1.0, 2.0, 1.0), Transposed.of(0.0, 0.0, 1.0)))
        );
        assertEquals(Transposed.sum(List.of(Transposed.of(1.0, 2.0, 1.0), Transposed.of(0.0, 0.0, 1.0))), Transposed.of(1.0, 2.0, 2.0));
    }

    @Test
    void transposedCreation() {
        assertEquals(Transposed.zeros(3), Transposed.of(0.0, 0.0, 0.0));
        assertEquals(Transposed.ones(3), Transposed.of(1.0, 1.0, 1.0));
        assertEquals(Transposed.of(1.0, 2.0).map( d -> d * 2.0), Transposed.of(2.0, 4.0));
        assertEquals(Transposed.fill(3, 1.7), Transposed.of(1.7, 1.7, 1.7));
        assertEquals(Transposed.tabulate(4, i -> i*5.0), Transposed.of(0.0, 5.0, 10.0, 15.0));
    }

    @Test
    void transposedConcat() {
        Transposed v1 = Transposed.of(1.0, 2.0, 3.0);
        Transposed v2 = Transposed.of(-1.0, -2.0, -3.0);
        assertEquals(v1.concat(v2), Transposed.of(1.0, 2.0, 3.0, -1.0, -2.0, -3.0));
    }

    @Test
    void transposedSlice() {
        Transposed v = Transposed.of(1.0, 2.0, 3.0, 4.0, 5.0);
        assertEquals(v.sliceFrom(2), Transposed.of(3.0, 4.0, 5.0));
        assertEquals(v.sliceTo(3), Transposed.of(1.0, 2.0, 3.0));
        assertEquals(v.slice(2,4), Transposed.of(3.0, 4.0));
        assertEquals(v.slice(4,4), Transposed.empty());
        assertEquals(v.slice(4,2), Transposed.empty());
        assertEquals(v.removed(2), Transposed.of(1.0, 2.0, 4.0, 5.0));
    }
// Vector V2 functions test :
    @Test
    void vectorfillv2() {
        assertEquals(Vector.fill(3).withValue(1.7), Vector.of(1.7, 1.7, 1.7));
        assertEquals(Vector.fill(4).withValue(1.5), Vector.of(1.5, 1.5, 1.5, 1.5));
        assertEquals(Vector.fill(2).withValue(1.0), Vector.of(1.0, 1.0));
    }

    @Test
    void vectorLinespacev2() {
        assertEquals(Vector.from(0.0).to(1.0).repeat(3), Vector.of(0.0, 0.5, 1.0));
        assertEquals(Vector.from(0.0).to(1.0).repeat(5), Vector.of(0.0, 0.25, 0.5, 0.75, 1.0));
    }
// Matrix
@Test
    void matrixCreation() {
        List<Vector> lstVec = new ArrayList<>();
        lstVec.add(Vector.of(1.0,1.0,1.0));
        lstVec.add(Vector.of(1.0,1.0,1.0));
        lstVec.add(Vector.of(1.0,1.0,1.0));
        Matrix mat = Matrix.of(lstVec);
        assertEquals(Matrix.of(lstVec), Matrix.of(lstVec));
        assertEquals(Matrix.of(lstVec), Matrix.of(lstVec));
        assertEquals(Matrix.of(lstVec).nbCols(), 3);
        assertEquals(Matrix.of(lstVec).nbRows(), 3);
    }

    @Test
    void matrixRemoved() {
        List<Vector> lstVec = new ArrayList<>();
        lstVec.add(Vector.of(1.0,2.0,3.0));
        lstVec.add(Vector.of(4.0,5.0,6.0));
        lstVec.add(Vector.of(7.0,8.0,9.0));
        Matrix mat = Matrix.of(lstVec);
        lstVec = new ArrayList<>();
        lstVec.add(Vector.of(5.0,6.0));
        lstVec.add(Vector.of(8.0,9.0));
        assertEquals(mat.removed(0, 0), Matrix.of(lstVec));
        
    }


    @Test
    void matrixMul() {
        List<Vector> lstVec = new ArrayList<>();
        lstVec.add(Vector.of(1.0,4.0,7.0));
        lstVec.add(Vector.of(2.0,5.0,8.0));
        lstVec.add(Vector.of(3.0,6.0,9.0));
        Matrix mat = Matrix.of(lstVec);
        lstVec = new ArrayList<>();
        lstVec.add(Vector.of(2.0,8.0,14.0));
        lstVec.add(Vector.of(4.0,10.0,16.0));
        lstVec.add(Vector.of(6.0,12.0,18.0));
        assertEquals(mat.mul(2.0), Matrix.of(lstVec)); 
        lstVec = new ArrayList<>();
        lstVec.add(Vector.of(3.0,12.0,21.0));
        lstVec.add(Vector.of(6.0,15.0,24.0));
        lstVec.add(Vector.of(9.0,18.0,27.0));
        assertEquals(mat.mul(3.0), Matrix.of(lstVec));
    }
    @Test
    void matrixDet() {
        List<Vector> lstVec = new ArrayList<>();
        lstVec.add(Vector.of(1.0,4.0,7.0));
        lstVec.add(Vector.of(2.0,5.0,8.0));
        lstVec.add(Vector.of(3.0,6.0,9.0));
        Matrix mat = Matrix.of(lstVec);
        assertEquals(mat.det(), 0.0);
        lstVec = new ArrayList<>();
        lstVec.add(Vector.of(1.0, 5.0, 9.0, 13.0));
        lstVec.add(Vector.of(2.0, 6.0, 10.0, 14.0));
        lstVec.add(Vector.of(3.0, 7.0, 11.0, 15.0));
        lstVec.add(Vector.of(4.0,8.0,12.0, 16.0));
        assertEquals(Matrix.of(lstVec).det(), 0.0);
    }

    @Test
    void matrixSet() {
        List<Vector> lstVec = new ArrayList<>();
        lstVec.add(Vector.of(1.0,4.0,7.0));
        lstVec.add(Vector.of(2.0,5.0,8.0));
        lstVec.add(Vector.of(3.0,6.0,9.0));
        Matrix mat = Matrix.of(lstVec);
        mat.set(1, 1, 10.0);
        lstVec = new ArrayList<>();
        lstVec.add(Vector.of(1.0,4.0,7.0));
        lstVec.add(Vector.of(2.0,10.0,8.0));
        lstVec.add(Vector.of(3.0,6.0,9.0));
        assertEquals(mat.cols().get(1).get(1),10.0);
    }

    @Test
    void matrixTabulate() {
        List<Vector> lstVec = new ArrayList<>();
        lstVec.add(Vector.of(0.0, 1.0, 2.0));
        lstVec.add(Vector.of(1.0, 2.0, 3.0));
        Matrix m = Matrix.of(lstVec);
        assertEquals(m, Matrix.tabulate(3, 2, (i, j) -> {
            return (double)i + j;
        }));
    }

    @Test
    void matrixInv() {
        List<Vector> lstVec = new ArrayList<>();
        lstVec.add(Vector.of(1.0, 1.0, 2.0));
        lstVec.add(Vector.of(1.0, 2.0, 1.0));
        lstVec.add(Vector.of(2.0, 1.0, 1.0));
        Matrix m = Matrix.of(lstVec);
        lstVec = new ArrayList<>();
        lstVec.add(Vector.of(-0.25, -0.25, 0.75));
        lstVec.add(Vector.of(-0.25, 0.75, -0.25));
        lstVec.add(Vector.of(0.75, -0.25, -0.25));
        assertEquals(m.inv(), Matrix.of(lstVec));

    }
}
